from .traj_opt_resolver import TrajOptResolver
from .db_resolver import DBResolver
import numpy as np

class TrajOptDBResolver(TrajOptResolver):
        
    def __init__(self, cell_size, grid_size, all_conflicts, all_robots, 
                 dt, robot_radius, starts, goals, circle_obstacles, 
                 rectangle_obstacles, rob_dist_weight, obs_dist_weight, time_weight):
        """
        inputs:
            - cell_size (float): size (height and width) of the cells to be placed in continuous space   
            - grid_size (int): size of the discrete grid we will place on top of continuous space. (nxn grid)  
            - robots_in_conflict (list): the indices of the robots that are in conflict
            - all_robots (list): list of all robots       
        """
        super.init(all_conflicts, all_robots, dt, robot_radius, starts, goals, circle_obstacles, 
                 rectangle_obstacles, rob_dist_weight, obs_dist_weight, time_weight)
        self.cell_size = cell_size
        self.grid_size = grid_size
        self.all_conflicts = all_conflicts
        self.all_robots = all_robots
        self.top_left_grid = None
        self.grid = None

    def place_grid(self, robot_locations):
        """
        Given the locations of robots that need to be covered in continuous space, find 
        and place the grid. We don't need a very large grid to place subproblems, so 
        we will only place a grid of size self.grid_size x self.grid_size

        inputs:
            - robot_locations (list): locations of robots involved in conflict
        outputs:
            - grid (numpy array): The grid that we placed
            - top_left (tuple): The top left corner of the grid in continuous space
        """
        # Find the minimum and maximum x and y coordinates of the robot locations
        min_x = min(robot_locations, key=lambda loc: loc[0])[0]
        max_x = max(robot_locations, key=lambda loc: loc[0])[0]
        min_y = min(robot_locations, key=lambda loc: loc[1])[1]
        max_y = max(robot_locations, key=lambda loc: loc[1])[1]

        # Calculate the top left corner of the grid
        self.top_left_grid = (min_x - .5*self.cell_size, max_y + .5*self.cell_size)

        # Initialize the grid
        grid = np.zeros((self.grid_size, self.grid_size), dtype=int)

        # Place robots in the grid
        for location in robot_locations:
            x, y = location
            cell_x = min(max(int((x - min_x) // self.cell_size), 0), self.grid_size - 1)
            cell_y = min(max(int((y - min_y) // self.cell_size), 0), self.grid_size - 1)
            grid[cell_x, cell_y] += 1

        # Check for conflicts and shift the grid if necessary
        for i in range(self.grid_size):
            for j in range(self.grid_size):
                if grid[i, j] > 1:
                    # Shift the grid by a small amount
                    shift = np.random.uniform(-0.1, 0.1, size=2)
                    grid[i, j] -= 1
                    new_i = min(max(i + int(shift[0]), 0), self.grid_size - 1)
                    new_j = min(max(j + int(shift[1]), 0), self.grid_size - 1)
                    grid[new_i, new_j] += 1

        return grid
    
    def get_discrete_solution(self, grid):
        #  create an instance of a discrete solver
        grid_solver = DBResolver(grid, self.conflict, self.all_robots, self.all_conflicts)
        subproblem = grid_solver.find_subproblem([],True,True)
        print(f"subproblem = {subproblem}")
        grid_solution = grid_solver.solve_subproblem(subproblem)
        return grid_solution
    
    def get_continuous_solution(self, grid_solution):
        raise NotImplementedError
        
    def get_local_controls(self):
        for c in self.conflicts:
            # Get the robots involved in the conflict
            robots = [self.all_robots[r.label] for r in c]
            robot_positions = [r.current_position for r in robots]

            # Put down a local grid
            self.grid = self.place_grid(robot_positions)

            # Find a subproblem and solve it
            grid_solution = self.get_discrete_solution(self.grid)

            # Solve the trajectory optimization problem, using the grid 
            # solution as an initial guess
            sol, x_opt = super.solve(10, grid_solution)

            # Update the controls for the robots
            for r, pos in zip(robots, x_opt):
                r.next_control = r.tracker.get_next_control(pos)
